# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorBoard helper routine for TF op evaluator.

Requires TensorFlow.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading


class PersistentOpEvaluator(object):
  """Evaluate a fixed TensorFlow graph repeatedly, safely, efficiently.

  Extend this class to create a particular kind of op evaluator, like an
  image encoder. In `initialize_graph`, create an appropriate TensorFlow
  graph with placeholder inputs. In `run`, evaluate this graph and
  return its result. This class will manage a singleton graph and
  session to preserve memory usage, and will ensure that this graph and
  session do not interfere with other concurrent sessions.

  A subclass of this class offers a threadsafe, highly parallel Python
  entry point for evaluating a particular TensorFlow graph.

  Example usage:

      class FluxCapacitanceEvaluator(PersistentOpEvaluator):
        \"\"\"Compute the flux capacitance required for a system.

        Arguments:
          x: Available power input, as a `float`, in jigawatts.

        Returns:
          A `float`, in nanofarads.
        \"\"\"

        def initialize_graph(self):
          self._placeholder = tf.placeholder(some_dtype)
          self._op = some_op(self._placeholder)

        def run(self, x):
          return self._op.eval(feed_dict: {self._placeholder: x})

      evaluate_flux_capacitance = FluxCapacitanceEvaluator()

      for x in xs:
        evaluate_flux_capacitance(x)
  """

  def __init__(self):
    super(PersistentOpEvaluator, self).__init__()
    self._session = None
    self._initialization_lock = threading.Lock()

  def _lazily_initialize(self):
    """Initialize the graph and session, if this has not yet been done."""
    # TODO(nickfelt): remove on-demand imports once dep situation is fixed.
    import tensorflow.compat.v1 as tf
    with self._initialization_lock:
      if self._session:
        return
      graph = tf.Graph()
      with graph.as_default():
        self.initialize_graph()
      # Don't reserve GPU because libpng can't run on GPU.
      config = tf.ConfigProto(device_count={'GPU': 0})
      self._session = tf.Session(graph=graph, config=config)

  def initialize_graph(self):
    """Create the TensorFlow graph needed to compute this operation.

    This should write ops to the default graph and return `None`.
    """
    raise NotImplementedError('Subclasses must implement "initialize_graph".')

  def run(self, *args, **kwargs):
    """Evaluate the ops with the given input.

    When this function is called, the default session will have the
    graph defined by a previous call to `initialize_graph`. This
    function should evaluate any ops necessary to compute the result of
    the query for the given *args and **kwargs, likely returning the
    result of a call to `some_op.eval(...)`.
    """
    raise NotImplementedError('Subclasses must implement "run".')

  def __call__(self, *args, **kwargs):
    self._lazily_initialize()
    with self._session.as_default():
      return self.run(*args, **kwargs)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper around plugin to conditionally enable it."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.plugins import base_plugin


class ProfilePluginLoader(base_plugin.TBLoader):
  """ProfilePlugin factory.

  This class checks for `tensorflow` install and dependency.
  """

  def define_flags(self, parser):
    group = parser.add_argument_group('profile plugin')
    group.add_argument(
        '--master_tpu_unsecure_channel',
        metavar='ADDR',
        type=str,
        default='',
        help='''\
IP address of "master tpu", used for getting streaming trace data
through tpu profiler analysis grpc. The grpc channel is not secured.\
''')

  def load(self, context):
    """Returns the plugin, if possible.

    Args:
      context: The TBContext flags.

    Returns:
      A ProfilePlugin instance or None if it couldn't be loaded.
    """
    try:
      # pylint: disable=g-import-not-at-top,unused-import
      import tensorflow
      # Available in TensorFlow 1.14 or later, so do import check
      # pylint: disable=g-import-not-at-top,unused-import
      from tensorflow.python.eager import profiler_client
    except ImportError:
      return
    # pylint: disable=g-import-not-at-top
    from tensorboard.plugins.profile.profile_plugin import ProfilePlugin
    return ProfilePlugin(context)



# -*- coding: utf-8 -*-
# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for handling Keras model in graph plugin.

Two canonical types of Keras model are Functional and Sequential.
A model can be serialized as JSON and deserialized to reconstruct a model.
This utility helps with dealing with the serialized Keras model.

They have distinct structures to the configurations in shapes below:
Functional:
  config
    name: Name of the model. If not specified, it is 'model' with
          an optional suffix if there are more than one instance.
    input_layers: Keras.layers.Inputs in the model.
    output_layers: Layer names that are outputs of the model.
    layers: list of layer configurations.
      layer: [*]
        inbound_nodes: inputs to this layer.

Sequential:
  config
    name: Name of the model. If not specified, it is 'sequential' with
          an optional suffix if there are more than one instance.
    layers: list of layer configurations.
      layer: [*]

[*]: Note that a model can be a layer.
Please refer to https://github.com/tensorflow/tfjs-layers/blob/master/src/keras_format/model_serialization.ts
for more complete definition.
"""
from tensorboard.compat.proto.graph_pb2 import GraphDef
from tensorboard.compat.tensorflow_stub import dtypes


def _walk_layers(keras_layer):
  """Walks the nested keras layer configuration in preorder.
  Args:
    keras_layer: Keras configuration from model.to_json.

  Yields:
    A tuple of (name_scope, layer_config).
    name_scope: a string representing a scope name, similar to that of tf.name_scope.
    layer_config: a dict representing a Keras layer configuration.
  """
  yield ('', keras_layer)
  if keras_layer.get('config').get('layers'):
    name_scope = keras_layer.get('config').get('name')
    for layer in keras_layer.get('config').get('layers'):
      for (sub_name_scope, sublayer) in _walk_layers(layer):
        sub_name_scope = '%s/%s' % (
            name_scope, sub_name_scope) if sub_name_scope else name_scope
        yield (sub_name_scope, sublayer)


def _scoped_name(name_scope, node_name):
  """Returns scoped name for a node as a string in the form '<scope>/<node name>'.

  Args:
    name_scope: a string representing a scope name, similar to that of tf.name_scope.
    node_name: a string representing the current node name.

  Returns
    A string representing a scoped name.
  """
  if name_scope:
    return '%s/%s' % (name_scope, node_name)
  return node_name


def _is_model(layer):
  """Returns True if layer is a model.

  Args:
    layer: a dict representing a Keras model configuration.

  Returns:
    bool: True if layer is a model.
  """
  return layer.get('config').get('layers') is not None


def _norm_to_list_of_layers(maybe_layers):
  """Normalizes to a list of layers.

  Args:
    maybe_layers: A list of data[1] or a list of list of data.

  Returns:
    List of list of data.

  [1]: A Functional model has fields 'inbound_nodes' and 'output_layers' which can
  look like below:
  - ['in_layer_name', 0, 0]
  - [['in_layer_is_model', 1, 0], ['in_layer_is_model', 1, 1]]
  The data inside the list seems to describe [name, size, index].
  """
  return (maybe_layers if isinstance(maybe_layers[0], (list,))
          else [maybe_layers])


def _update_dicts(name_scope,
                  model_layer,
                  input_to_in_layer,
                  model_name_to_output,
                  prev_node_name):
  """Updates input_to_in_layer, model_name_to_output, and prev_node_name
  based on the model_layer.

  Args:
    name_scope: a string representing a scope name, similar to that of tf.name_scope.
    model_layer: a dict representing a Keras model configuration.
    input_to_in_layer: a dict mapping Keras.layers.Input to inbound layer.
    model_name_to_output: a dict mapping Keras Model name to output layer of the model.
    prev_node_name: a string representing a previous, in sequential model layout,
                    node name.

  Returns:
    A tuple of (input_to_in_layer, model_name_to_output, prev_node_name).
    input_to_in_layer: a dict mapping Keras.layers.Input to inbound layer.
    model_name_to_output: a dict mapping Keras Model name to output layer of the model.
    prev_node_name: a string representing a previous, in sequential model layout,
                    node name.
  """
  layer_config = model_layer.get('config')
  if not layer_config.get('layers'):
    raise ValueError('layer is not a model.')

  node_name = _scoped_name(name_scope, layer_config.get('name'))
  input_layers = layer_config.get('input_layers')
  output_layers = layer_config.get('output_layers')
  inbound_nodes = model_layer.get('inbound_nodes')

  is_functional_model = bool(input_layers and output_layers)
  # In case of [1] and the parent model is functional, current layer
  # will have the 'inbound_nodes' property.
  is_parent_functional_model = bool(inbound_nodes)

  if is_parent_functional_model and is_functional_model:
    for (input_layer, inbound_node) in zip(input_layers, inbound_nodes):
      input_layer_name = _scoped_name(node_name, input_layer)
      inbound_node_name = _scoped_name(name_scope, inbound_node[0])
      input_to_in_layer[input_layer_name] = inbound_node_name
  elif is_parent_functional_model and not is_functional_model:
    # Sequential model can take only one input. Make sure inbound to the
    # model is linked to the first layer in the Sequential model.
    prev_node_name = _scoped_name(name_scope, inbound_nodes[0][0][0])
  elif not is_parent_functional_model and prev_node_name and is_functional_model:
    assert len(input_layers) == 1, (
        'Cannot have multi-input Functional model when parent model '
        'is not Functional. Number of input layers: %d' % len(input_layer))
    input_layer = input_layers[0]
    input_layer_name = _scoped_name(node_name, input_layer)
    input_to_in_layer[input_layer_name] = prev_node_name

  if is_functional_model and output_layers:
    layers = _norm_to_list_of_layers(output_layers)
    layer_names = [_scoped_name(node_name, layer[0]) for layer in layers]
    model_name_to_output[node_name] = layer_names
  else:
    last_layer = layer_config.get('layers')[-1]
    last_layer_name = last_layer.get('config').get('name')
    output_node = _scoped_name(node_name, last_layer_name)
    model_name_to_output[node_name] = [output_node]
  return (input_to_in_layer, model_name_to_output, prev_node_name)


def keras_model_to_graph_def(keras_layer):
  """Returns a GraphDef representation of the Keras model in a dict form.

  Note that it only supports models that implemented to_json().

  Args:
    keras_layer: A dict from Keras model.to_json().

  Returns:
    A GraphDef representation of the layers in the model.
  """
  input_to_layer = {}
  model_name_to_output = {}
  g = GraphDef()

  # Sequential model layers do not have a field "inbound_nodes" but
  # instead are defined implicitly via order of layers.
  prev_node_name = None

  for (name_scope, layer) in _walk_layers(keras_layer):
    if _is_model(layer):
      (input_to_layer, model_name_to_output, prev_node_name) = _update_dicts(
          name_scope, layer, input_to_layer, model_name_to_output, prev_node_name)
      continue

    layer_config = layer.get('config')
    node_name = _scoped_name(name_scope, layer_config.get('name'))

    node_def = g.node.add()
    node_def.name = node_name

    if layer.get('class_name') is not None:
      keras_cls_name = layer.get('class_name').encode('ascii')
      node_def.attr['keras_class'].s = keras_cls_name

    if layer_config.get('dtype') is not None:
      tf_dtype = dtypes.as_dtype(layer_config.get('dtype'))
      node_def.attr['dtype'].type = tf_dtype.as_datatype_enum

    if layer.get('inbound_nodes') is not None:
      for maybe_inbound_node in layer.get('inbound_nodes'):
        inbound_nodes = _norm_to_list_of_layers(maybe_inbound_node)
        for [name, size, index, _] in inbound_nodes:
          inbound_name = _scoped_name(name_scope, name)
          # An input to a layer can be output from a model. In that case, the name
          # of inbound_nodes to a layer is a name of a model. Remap the name of the
          # model to output layer of the model. Also, since there can be multiple
          # outputs in a model, make sure we pick the right output_layer from the model.
          inbound_node_names = model_name_to_output.get(
              inbound_name, [inbound_name])
          node_def.input.append(inbound_node_names[index])
    elif prev_node_name is not None:
      node_def.input.append(prev_node_name)

    if node_name in input_to_layer:
      node_def.input.append(input_to_layer.get(node_name))

    prev_node_name = node_def.name

  return g



# -*- coding: utf-8 -*-
# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""The TensorBoard plugin for performance profiling."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import threading

import six
import tensorflow as tf
from werkzeug import wrappers

from tensorflow.python.eager import profiler_client
from tensorboard.backend import http_util
from tensorboard.backend.event_processing import plugin_asset_util
from tensorboard.plugins import base_plugin
from tensorboard.plugins.profile import trace_events_json
from tensorboard.plugins.profile import trace_events_pb2
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

# The prefix of routes provided by this plugin.
PLUGIN_NAME = 'profile'

# HTTP routes
DATA_ROUTE = '/data'
TOOLS_ROUTE = '/tools'
HOSTS_ROUTE = '/hosts'
CAPTURE_ROUTE = '/capture_profile'

# Available profiling tools -> file name of the tool data.
_FILE_NAME = 'TOOL_FILE_NAME'
TOOLS = {
    'trace_viewer': 'trace',
    'trace_viewer@': 'tracetable',  #streaming traceviewer
    'op_profile': 'op_profile.json',
    'input_pipeline_analyzer': 'input_pipeline.json',
    'overview_page': 'overview_page.json',
    'memory_viewer': 'memory_viewer.json',
    'pod_viewer': 'pod_viewer.json',
    'google_chart_demo': 'google_chart_demo.json',
}

# Tools that consume raw data.
_RAW_DATA_TOOLS = frozenset(['input_pipeline_analyzer',
                             'op_profile',
                             'overview_page',
                             'memory_viewer',
                             'pod_viewer',
                             'google_chart_demo',])

def process_raw_trace(raw_trace):
  """Processes raw trace data and returns the UI data."""
  trace = trace_events_pb2.Trace()
  trace.ParseFromString(raw_trace)
  return ''.join(trace_events_json.TraceEventsJsonStream(trace))

def get_worker_list(cluster_resolver):
  """Parses TPU workers list from the cluster resolver."""
  cluster_spec = cluster_resolver.cluster_spec()
  task_indices = cluster_spec.task_indices('worker')
  worker_list = [
      cluster_spec.task_address('worker', i).split(':')[0] for i in task_indices
  ]
  return ','.join(worker_list)

class ProfilePlugin(base_plugin.TBPlugin):
  """Profile Plugin for TensorBoard."""

  plugin_name = PLUGIN_NAME

  def __init__(self, context):
    """Constructs a profiler plugin for TensorBoard.

    This plugin adds handlers for performance-related frontends.

    Args:
      context: A base_plugin.TBContext instance.
    """
    self.logdir = context.logdir
    self.multiplexer = context.multiplexer
    self.plugin_logdir = plugin_asset_util.PluginDirectory(
        self.logdir, PLUGIN_NAME)
    self.stub = None
    self.master_tpu_unsecure_channel = context.flags.master_tpu_unsecure_channel

    # Whether the plugin is active. This is an expensive computation, so we
    # compute this asynchronously and cache positive results indefinitely.
    self._is_active = False
    # Lock to ensure at most one thread computes _is_active at a time.
    self._is_active_lock = threading.Lock()

  def is_active(self):
    """Whether this plugin is active and has any profile data to show.

    Detecting profile data is expensive, so this process runs asynchronously
    and the value reported by this method is the cached value and may be stale.

    Returns:
      Whether any run has profile data.
    """
    # If we are already active, we remain active and don't recompute this.
    # Otherwise, try to acquire the lock without blocking; if we get it and
    # we're still not active, launch a thread to check if we're active and
    # release the lock once the computation is finished. Either way, this
    # thread returns the current cached value to avoid blocking.
    if not self._is_active and self._is_active_lock.acquire(False):
      if self._is_active:
        self._is_active_lock.release()
      else:
        def compute_is_active():
          self._is_active = any(self.generate_run_to_tools())
          self._is_active_lock.release()
        new_thread = threading.Thread(
            target=compute_is_active,
            name='ProfilePluginIsActiveThread')
        new_thread.start()
    return self._is_active

  def frontend_metadata(self):
    # TODO(#2338): Keep this in sync with the `registerDashboard` call
    # on the frontend until that call is removed.
    return base_plugin.FrontendMetadata(
        element_name='tf-profile-dashboard',
        disable_reload=True,
    )

  def start_grpc_stub_if_necessary(self):
    # We will enable streaming trace viewer on two conditions:
    # 1. user specify the flags master_tpu_unsecure_channel to the ip address of
    #    as "master" TPU. grpc will be used to fetch streaming trace data.
    # 2. the logdir is on google cloud storage.
    if self.master_tpu_unsecure_channel and self.logdir.startswith('gs://'):
      if self.stub is None:
        import grpc
        from tensorflow.python.tpu.profiler import profiler_analysis_pb2_grpc # pylint: disable=line-too-long
        # Workaround the grpc's 4MB message limitation.
        gigabyte = 1024 * 1024 * 1024
        options = [('grpc.max_message_length', gigabyte),
                   ('grpc.max_send_message_length', gigabyte),
                   ('grpc.max_receive_message_length', gigabyte)]
        tpu_profiler_port = self.master_tpu_unsecure_channel + ':8466'
        channel = grpc.insecure_channel(tpu_profiler_port, options)
        self.stub = profiler_analysis_pb2_grpc.ProfileAnalysisStub(channel)

  def _run_dir(self, run):
    """Helper that maps a frontend run name to a profile "run" directory.

    The frontend run name consists of the TensorBoard run name (aka the relative
    path from the logdir root to the directory containing the data) path-joined
    to the Profile plugin's "run" concept (which is a subdirectory of the
    plugins/profile directory representing an individual run of the tool), with
    the special case that TensorBoard run is the logdir root (which is the run
    named '.') then only the Profile plugin "run" name is used, for backwards
    compatibility.

    To convert back to the actual run directory, we apply the following
    transformation:
    - If the run name doesn't contain '/', prepend './'
    - Split on the rightmost instance of '/'
    - Assume the left side is a TensorBoard run name and map it to a directory
      path using EventMultiplexer.RunPaths(), then map that to the profile
      plugin directory via PluginDirectory()
    - Assume the right side is a Profile plugin "run" and path-join it to
      the preceding path to get the final directory

    Args:
      run: the frontend run name, as described above, e.g. train/run1.

    Returns:
      The resolved directory path, e.g. /logdir/train/plugins/profile/run1.
    """
    run = run.rstrip('/')
    if '/' not in run:
      run = './' + run
    tb_run_name, _, profile_run_name = run.rpartition('/')
    tb_run_directory = self.multiplexer.RunPaths().get(tb_run_name)
    if tb_run_directory is None:
      # Check if logdir is a directory to handle case where it's actually a
      # multipart directory spec, which this plugin does not support.
      if tb_run_name == '.' and tf.io.gfile.isdir(self.logdir):
        tb_run_directory = self.logdir
      else:
        raise RuntimeError("No matching run directory for run %s" % run)
    plugin_directory = plugin_asset_util.PluginDirectory(
        tb_run_directory, PLUGIN_NAME)
    return os.path.join(plugin_directory, profile_run_name)

  def generate_run_to_tools(self):
    """Generator for pairs of "run name" and a list of tools for that run.

    The "run name" here is a "frontend run name" - see _run_dir() for the
    definition of a "frontend run name" and how it maps to a directory of
    profile data for a specific profile "run". The profile plugin concept of
    "run" is different from the normal TensorBoard run; each run in this case
    represents a single instance of profile data collection, more similar to a
    "step" of data in typical TensorBoard semantics. These runs reside in
    subdirectories of the plugins/profile directory within any regular
    TensorBoard run directory (defined as a subdirectory of the logdir that
    contains at least one tfevents file) or within the logdir root directory
    itself (even if it contains no tfevents file and would thus not be
    considered a normal TensorBoard run, for backwards compatibility).

    Within those "profile run directories", there are files in the directory
    that correspond to different profiling tools. The file that contains profile
    for a specific tool "x" will have a suffix name TOOLS["x"].

    Example:
      logs/
        plugins/
          profile/
            run1/
              hostA.trace
        train/
          events.out.tfevents.foo
          plugins/
            profile/
              run1/
                hostA.trace
                hostB.trace
              run2/
                hostA.trace
        validation/
          events.out.tfevents.foo
          plugins/
            profile/
              run1/
                hostA.trace

    Yields:
      A sequence of tuples mapping "frontend run names" to lists of tool names
      available for those runs. For the above example, this would be:

          ("run1", ["trace_viewer"])
          ("train/run1", ["trace_viewer"])
          ("train/run2", ["trace_viewer"])
          ("validation/run1", ["trace_viewer"])
    """
    self.start_grpc_stub_if_necessary()

    plugin_assets = self.multiplexer.PluginAssets(PLUGIN_NAME)
    tb_run_names_to_dirs = self.multiplexer.RunPaths()

    # Ensure that we also check the root logdir, even if it isn't a recognized
    # TensorBoard run (i.e. has no tfevents file directly under it), to remain
    # backwards compatible with previously profile plugin behavior. Note that we
    # check if logdir is a directory to handle case where it's actually a
    # multipart directory spec, which this plugin does not support.
    if '.' not in plugin_assets and tf.io.gfile.isdir(self.logdir):
      tb_run_names_to_dirs['.'] = self.logdir
      plugin_assets['.'] = plugin_asset_util.ListAssets(
          self.logdir, PLUGIN_NAME)

    for tb_run_name, profile_runs in six.iteritems(plugin_assets):
      tb_run_dir = tb_run_names_to_dirs[tb_run_name]
      tb_plugin_dir = plugin_asset_util.PluginDirectory(
          tb_run_dir, PLUGIN_NAME)
      for profile_run in profile_runs:
        # Remove trailing slash; some filesystem implementations emit this.
        profile_run = profile_run.rstrip('/')
        if tb_run_name == '.':
          frontend_run = profile_run
        else:
          frontend_run = '/'.join([tb_run_name, profile_run])
        profile_run_dir = os.path.join(tb_plugin_dir, profile_run)
        if tf.io.gfile.isdir(profile_run_dir):
          yield frontend_run, self._get_active_tools(profile_run_dir)

  def _get_active_tools(self, profile_run_dir):
    tools = []
    for tool in TOOLS:
      tool_pattern = '*' + TOOLS[tool]
      path = os.path.join(profile_run_dir, tool_pattern)
      try:
        files = tf.io.gfile.glob(path)
        if len(files) >= 1:
          tools.append(tool)
      except tf.errors.OpError as e:
        logger.warn("Cannot read asset directory: %s, OpError %s",
                    profile_run_dir, e)
    if 'trace_viewer@' in tools:
      # streaming trace viewer always override normal trace viewer.
      # the trailing '@' is to inform tf-profile-dashboard.html and
      # tf-trace-viewer.html that stream trace viewer should be used.
      removed_tool = 'trace_viewer@' if self.stub is None else 'trace_viewer'
      if removed_tool in tools:
        tools.remove(removed_tool)
    tools.sort()
    op = 'overview_page'
    if op in tools:
      # keep overview page at the top of the list
      tools.remove(op)
      tools.insert(0, op)
    return tools

  @wrappers.Request.application
  def tools_route(self, request):
    run_to_tools = dict(self.generate_run_to_tools())
    return http_util.Respond(request, run_to_tools, 'application/json')

  def host_impl(self, run, tool):
    """Returns available hosts for the run and tool in the log directory.

    In the plugin log directory, each directory contains profile data for a
    single run (identified by the directory name), and files in the run
    directory contains data for different tools and hosts. The file that
    contains profile for a specific tool "x" will have a prefix name TOOLS["x"].

    Example:
      log/
        run1/
          plugins/
            profile/
              host1.trace
              host2.trace
        run2/
          plugins/
            profile/
              host1.trace
              host2.trace

    Returns:
      A list of host names e.g.
        {"host1", "host2", "host3"} for the example.
    """
    hosts = {}
    run_dir = self._run_dir(run)
    if not run_dir:
      logger.warn("Cannot find asset directory for: %s", run)
      return hosts
    tool_pattern = '*' + TOOLS[tool]
    try:
      files = tf.io.gfile.glob(os.path.join(run_dir, tool_pattern))
      hosts = [os.path.basename(f).replace(TOOLS[tool], '') for f in files]
    except tf.errors.OpError as e:
      logger.warn("Cannot read asset directory: %s, OpError %s",
                      run_dir, e)
    return hosts


  @wrappers.Request.application
  def hosts_route(self, request):
    run = request.args.get('run')
    tool = request.args.get('tag')
    hosts = self.host_impl(run, tool)
    return http_util.Respond(request, hosts, 'application/json')

  def data_impl(self, request):
    """Retrieves and processes the tool data for a run and a host.

    Args:
      request: XMLHttpRequest

    Returns:
      A string that can be served to the frontend tool or None if tool,
        run or host is invalid.
    """
    run = request.args.get('run')
    tool = request.args.get('tag')
    host = request.args.get('host')
    run_dir = self._run_dir(run)
    # Profile plugin "run" is the last component of run dir.
    profile_run = os.path.basename(run_dir)

    if tool not in TOOLS:
      return None

    self.start_grpc_stub_if_necessary()
    if tool == 'trace_viewer@' and self.stub is not None:
      from tensorflow.core.profiler import profiler_analysis_pb2
      grpc_request = profiler_analysis_pb2.ProfileSessionDataRequest()
      grpc_request.repository_root = os.path.dirname(run_dir)
      grpc_request.session_id = profile_run
      grpc_request.tool_name = 'trace_viewer'
      # Remove the trailing dot if present
      grpc_request.host_name = host.rstrip('.')

      grpc_request.parameters['resolution'] = request.args.get('resolution')
      if request.args.get('start_time_ms') is not None:
        grpc_request.parameters['start_time_ms'] = request.args.get(
            'start_time_ms')
      if request.args.get('end_time_ms') is not None:
        grpc_request.parameters['end_time_ms'] = request.args.get('end_time_ms')
      grpc_response = self.stub.GetSessionToolData(grpc_request)
      return grpc_response.output

    if tool not in TOOLS:
      return None
    tool_name = str(host) + TOOLS[tool]
    asset_path = os.path.join(run_dir, tool_name)
    raw_data = None
    try:
      with tf.io.gfile.GFile(asset_path, 'rb') as f:
        raw_data = f.read()
    except tf.errors.NotFoundError:
      logger.warn('Asset path %s not found', asset_path)
    except tf.errors.OpError as e:
      logger.warn("Couldn't read asset path: %s, OpError %s", asset_path, e)

    if raw_data is None:
      return None
    if tool == 'trace_viewer':
      return process_raw_trace(raw_data)
    if tool in _RAW_DATA_TOOLS:
      return raw_data
    return None

  @wrappers.Request.application
  def data_route(self, request):
    # params
    #   request: XMLHTTPRequest.
    data = self.data_impl(request)
    if data is None:
      return http_util.Respond(request, '404 Not Found', 'text/plain', code=404)
    return http_util.Respond(request, data, 'application/json')

  @wrappers.Request.application
  def capture_route(self, request):
    service_addr = request.args.get('service_addr')
    duration = int(request.args.get('duration', '1000'))
    is_tpu_name = request.args.get('is_tpu_name') == 'true'
    worker_list = request.args.get('worker_list')
    include_dataset_ops = request.args.get('include_dataset_ops') == 'true'
    num_tracing_attempts = int(request.args.get('num_retry', '0')) + 1

    if is_tpu_name:
      try:
        tpu_cluster_resolver = (
            tf.distribute.cluster_resolver.TPUClusterResolver(service_addr))
        master_grpc_addr = tpu_cluster_resolver.get_master()
      except (ImportError, RuntimeError) as err:
        return http_util.Respond(request, {'error': err.message},
                                 'application/json', code=200)
      except (ValueError, TypeError):
        return http_util.Respond(request,
            {'error': 'no TPUs with the specified names exist.'},
             'application/json', code=200)
      if not worker_list:
        worker_list = get_worker_list(tpu_cluster_resolver)
      # TPU cluster resolver always returns port 8470. Replace it with 8466
      # on which profiler service is running.
      master_ip = master_grpc_addr.replace('grpc://', '').replace(':8470', '')
      service_addr = master_ip + ':8466'
      # Set the master TPU for streaming trace viewer.
      self.master_tpu_unsecure_channel = master_ip
    try:
      profiler_client.start_tracing(service_addr, self.logdir, duration,
          worker_list, include_dataset_ops, num_tracing_attempts)
      return http_util.Respond(
          request, {'result': 'Capture profile successfully. Please refresh.'},
          'application/json')
    except tf.errors.UnavailableError:
      return http_util.Respond(request, {'error': 'empty trace result.'},
                                         'application/json', code=200)

  def get_plugin_apps(self):
    return {
        TOOLS_ROUTE: self.tools_route,
        HOSTS_ROUTE: self.hosts_route,
        DATA_ROUTE: self.data_route,
        CAPTURE_ROUTE: self.capture_route,
    }



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Summary creation methods for the HParams plugin.

Typical usage for exporting summaries in a hyperparameters-tuning experiment:
1. Create the experiment (once) by calling experiment_pb() and exporting
   the resulting summary into a top-level (empty) run.
2. In each training session in the experiment, call session_start_pb() before
   the session starts, exporting the resulting summary into a uniquely named
   run for the session, say <session_name>.
3. Train the model in the session, exporting each metric as a scalar summary
   in runs of the form <session_name>/<sub_dir>, where <sub_dir> can be empty a
   (in which case the run is just the <session_name>) and depends on the
   metric. The name of such a metric is a (group, tag) pair given by
   (<sub_dir>, tag) where tag is the tag of the scalar summary.
   When calling experiment_pb in step 1, you'll need to pass all the metric
   names used in the experiemnt.
4. When the session completes, call session_end_pb() and export the resulting
   summary into the same session run <session_name>.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time

import six

import tensorflow as tf

from tensorboard.plugins.hparams import api_pb2
from tensorboard.plugins.hparams import metadata
from tensorboard.plugins.hparams import plugin_data_pb2


def experiment_pb(
    hparam_infos,
    metric_infos,
    user='',
    description='',
    time_created_secs=None):
  """Creates a summary that defines a hyperparameter-tuning experiment.

  Args:
    hparam_infos: Array of api_pb2.HParamInfo messages. Describes the
        hyperparameters used in the experiment.
    metric_infos: Array of api_pb2.MetricInfo messages. Describes the metrics
        used in the experiment. See the documentation at the top of this file
        for how to populate this.
    user: String. An id for the user running the experiment
    description: String. A description for the experiment. May contain markdown.
    time_created_secs: float. The time the experiment is created in seconds
    since the UNIX epoch. If None uses the current time.

  Returns:
    A summary protobuffer containing the experiment definition.
  """
  if time_created_secs is None:
    time_created_secs = time.time()
  experiment = api_pb2.Experiment(
      description=description,
      user=user,
      time_created_secs=time_created_secs,
      hparam_infos=hparam_infos,
      metric_infos=metric_infos)
  return _summary(metadata.EXPERIMENT_TAG,
                  plugin_data_pb2.HParamsPluginData(experiment=experiment))


def session_start_pb(hparams,
                     model_uri='',
                     monitor_url='',
                     group_name='',
                     start_time_secs=None):
  """Constructs a SessionStartInfo protobuffer.

  Creates a summary that contains a training session metadata information.
  One such summary per training session should be created. Each should have
  a different run.

  Args:
    hparams: A dictionary with string keys. Describes the hyperparameter values
             used in the session, mapping each hyperparameter name to its value.
             Supported value types are  `bool`, `int`, `float`, `str`, `list`,
             `tuple`.
             The type of value must correspond to the type of hyperparameter
             (defined in the corresponding api_pb2.HParamInfo member of the
             Experiment protobuf) as follows:

              +-----------------+---------------------------------+
              |Hyperparameter   | Allowed (Python) value types    |
              |type             |                                 |
              +-----------------+---------------------------------+
              |DATA_TYPE_BOOL   | bool                            |
              |DATA_TYPE_FLOAT64| int, float                      |
              |DATA_TYPE_STRING | six.string_types, tuple, list   |
              +-----------------+---------------------------------+

             Tuple and list instances will be converted to their string
             representation.
    model_uri: See the comment for the field with the same name of
               plugin_data_pb2.SessionStartInfo.
    monitor_url: See the comment for the field with the same name of
                 plugin_data_pb2.SessionStartInfo.
    group_name:  See the comment for the field with the same name of
                 plugin_data_pb2.SessionStartInfo.
    start_time_secs: float. The time to use as the session start time.
                     Represented as seconds since the UNIX epoch. If None uses
                     the current time.
  Returns:
    The summary protobuffer mentioned above.
  """
  if start_time_secs is None:
    start_time_secs = time.time()
  session_start_info = plugin_data_pb2.SessionStartInfo(
      model_uri=model_uri,
      monitor_url=monitor_url,
      group_name=group_name,
      start_time_secs=start_time_secs)
  for (hp_name, hp_val) in six.iteritems(hparams):
    if isinstance(hp_val, (float, int)):
      session_start_info.hparams[hp_name].number_value = hp_val
    elif isinstance(hp_val, six.string_types):
      session_start_info.hparams[hp_name].string_value = hp_val
    elif isinstance(hp_val, bool):
      session_start_info.hparams[hp_name].bool_value = hp_val
    elif isinstance(hp_val, (list, tuple)):
      session_start_info.hparams[hp_name].string_value = str(hp_val)
    else:
      raise TypeError('hparams[%s]=%s has type: %s which is not supported' %
                      (hp_name, hp_val, type(hp_val)))
  return _summary(metadata.SESSION_START_INFO_TAG,
                  plugin_data_pb2.HParamsPluginData(
                      session_start_info=session_start_info))


def session_end_pb(status, end_time_secs=None):
  """Constructs a SessionEndInfo protobuffer.

  Creates a summary that contains status information for a completed
  training session. Should be exported after the training session is completed.
  One such summary per training session should be created. Each should have
  a different run.

  Args:
    status: A tensorboard.hparams.Status enumeration value denoting the
        status of the session.
    end_time_secs: float. The time to use as the session end time. Represented
        as seconds since the unix epoch. If None uses the current time.

  Returns:
    The summary protobuffer mentioned above.
  """
  if end_time_secs is None:
    end_time_secs = time.time()

  session_end_info = plugin_data_pb2.SessionEndInfo(status=status,
                                                    end_time_secs=end_time_secs)
  return _summary(metadata.SESSION_END_INFO_TAG,
                  plugin_data_pb2.HParamsPluginData(
                      session_end_info=session_end_info))


def _summary(tag, hparams_plugin_data):
  """Returns a summary holding the given HParamsPluginData message.

  Helper function.

  Args:
    tag: string. The tag to use.
    hparams_plugin_data: The HParamsPluginData message to use.
  """
  summary = tf.compat.v1.Summary()
  tb_metadata = metadata.create_summary_metadata(hparams_plugin_data)
  raw_metadata = tb_metadata.SerializeToString()
  tf_metadata = tf.compat.v1.SummaryMetadata.FromString(raw_metadata)
  summary.value.add(tag=tag, metadata=tf_metadata)
  return summary



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Experimental framework for generic TensorBoard data providers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import collections

import six


@six.add_metaclass(abc.ABCMeta)
class DataProvider(object):
  """Interface for reading TensorBoard scalar, tensor, and blob data.

  These APIs are under development and subject to change. For instance,
  providers may be asked to implement more filtering mechanisms, such as
  downsampling strategies or domain restriction by step or wall time.

  Unless otherwise noted, any methods on this class may raise errors
  defined in `tensorboard.errors`, like `tensorboard.errors.NotFound`.
  """

  def data_location(self, experiment_id):
    """Render a human-readable description of the data source.

    For instance, this might return a path to a directory on disk.

    The default implementation always returns the empty string.

    Args:
      experiment_id: ID of enclosing experiment.

    Returns:
      A string, which may be empty.
    """
    return ""


  @abc.abstractmethod
  def list_runs(self, experiment_id):
    """List all runs within an experiment.

    Args:
      experiment_id: ID of enclosing experiment.

    Returns:
      A collection of `Run` values.

    Raises:
      tensorboard.errors.PublicError: See `DataProvider` class docstring.
    """
    pass

  @abc.abstractmethod
  def list_scalars(self, experiment_id, plugin_name, run_tag_filter=None):
    """List metadata about scalar time series.

    Args:
      experiment_id: ID of enclosing experiment.
      plugin_name: String name of the TensorBoard plugin that created
        the data to be queried. Required.
      run_tag_filter: Optional `RunTagFilter` value. If omitted, all
        runs and tags will be included.

    The result will only contain keys for run-tag combinations that
    actually exist, which may not include all entries in the
    `run_tag_filter`.

    Returns:
      A nested map `d` such that `d[run][tag]` is a `ScalarTimeSeries`
      value.

    Raises:
      tensorboard.errors.PublicError: See `DataProvider` class docstring.
    """
    pass

  @abc.abstractmethod
  def read_scalars(
      self, experiment_id, plugin_name, downsample=None, run_tag_filter=None
  ):
    """Read values from scalar time series.

    Args:
      experiment_id: ID of enclosing experiment.
      plugin_name: String name of the TensorBoard plugin that created
        the data to be queried. Required.
      downsample: Integer number of steps to which to downsample the
        results (e.g., `1000`). Required.
      run_tag_filter: Optional `RunTagFilter` value. If provided, a time
        series will only be included in the result if its run and tag
        both pass this filter. If `None`, all time series will be
        included.

    The result will only contain keys for run-tag combinations that
    actually exist, which may not include all entries in the
    `run_tag_filter`.

    Returns:
      A nested map `d` such that `d[run][tag]` is a list of
      `ScalarDatum` values sorted by step.

    Raises:
      tensorboard.errors.PublicError: See `DataProvider` class docstring.
    """
    pass

  def list_tensors(self):
    """Not yet specified."""
    pass

  def read_tensors(self):
    """Not yet specified."""
    pass

  def list_blob_sequences(self):
    """Not yet specified."""
    pass

  def read_blob_sequences(self):
    """Not yet specified."""
    pass


class Run(object):
  """Metadata about a run.

  Attributes:
    run_id: A unique opaque string identifier for this run.
    run_name: A user-facing name for this run (as a `str`).
    start_time: The wall time of the earliest recorded event in this
      run, as `float` seconds since epoch, or `None` if this run has no
      recorded events.
  """

  __slots__ = ("_run_id", "_run_name", "_start_time")

  def __init__(self, run_id, run_name, start_time):
    self._run_id = run_id
    self._run_name = run_name
    self._start_time = start_time

  @property
  def run_id(self):
    return self._run_id

  @property
  def run_name(self):
    return self._run_name

  @property
  def start_time(self):
    return self._start_time

  def __eq__(self, other):
    if not isinstance(other, Run):
      return False
    if self._run_id != other._run_id:
      return False
    if self._run_name != other._run_name:
      return False
    if self._start_time != other._start_time:
      return False
    return True

  def __hash__(self):
    return hash((self._run_id, self._run_name, self._start_time))

  def __repr__(self):
    return "Run(%s)" % ", ".join((
        "run_id=%r" % (self._run_id,),
        "run_name=%r" % (self._run_name,),
        "start_time=%r" % (self._start_time,),
    ))


class ScalarTimeSeries(object):
  """Metadata about a scalar time series for a particular run and tag.

  Attributes:
    max_step: The largest step value of any datum in this scalar time
      series; a nonnegative integer.
    max_wall_time: The largest wall time of any datum in this time
      series, as `float` seconds since epoch.
    plugin_content: A bytestring of arbitrary plugin-specific metadata
      for this time series, as provided to `tf.summary.write` in the
      `plugin_data.content` field of the `metadata` argument.
    description: An optional long-form Markdown description, as a `str`
      that is empty if no description was specified.
    display_name: An optional long-form Markdown description, as a `str`
      that is empty if no description was specified. Deprecated; may be
      removed soon.
  """

  __slots__ = (
      "_max_step",
      "_max_wall_time",
      "_plugin_content",
      "_description",
      "_display_name",
  )

  def __init__(
      self, max_step, max_wall_time, plugin_content, description, display_name
  ):
    self._max_step = max_step
    self._max_wall_time = max_wall_time
    self._plugin_content = plugin_content
    self._description = description
    self._display_name = display_name

  @property
  def max_step(self):
    return self._max_step

  @property
  def max_wall_time(self):
    return self._max_wall_time

  @property
  def plugin_content(self):
    return self._plugin_content

  @property
  def description(self):
    return self._description

  @property
  def display_name(self):
    return self._display_name

  def __eq__(self, other):
    if not isinstance(other, ScalarTimeSeries):
      return False
    if self._max_step != other._max_step:
      return False
    if self._max_wall_time != other._max_wall_time:
      return False
    if self._plugin_content != other._plugin_content:
      return False
    if self._description != other._description:
      return False
    if self._display_name != other._display_name:
      return False
    return True

  def __hash__(self):
    return hash((
        self._max_step,
        self._max_wall_time,
        self._plugin_content,
        self._description,
        self._display_name,
    ))

  def __repr__(self):
    return "ScalarTimeSeries(%s)" % ", ".join((
        "max_step=%r" % (self._max_step,),
        "max_wall_time=%r" % (self._max_wall_time,),
        "plugin_content=%r" % (self._plugin_content,),
        "description=%r" % (self._description,),
        "display_name=%r" % (self._display_name,),
    ))


class ScalarDatum(object):
  """A single datum in a scalar time series for a run and tag.

  Attributes:
    step: The global step at which this datum occurred; an integer. This
      is a unique key among data of this time series.
    wall_time: The real-world time at which this datum occurred, as
      `float` seconds since epoch.
    value: The scalar value for this datum; a `float`.
  """

  __slots__ = ("_step", "_wall_time", "_value")

  def __init__(self, step, wall_time, value):
    self._step = step
    self._wall_time = wall_time
    self._value = value

  @property
  def step(self):
    return self._step

  @property
  def wall_time(self):
    return self._wall_time

  @property
  def value(self):
    return self._value

  def __eq__(self, other):
    if not isinstance(other, ScalarDatum):
      return False
    if self._step != other._step:
      return False
    if self._wall_time != other._wall_time:
      return False
    if self._value != other._value:
      return False
    return True

  def __hash__(self):
    return hash((self._step, self._wall_time, self._value))

  def __repr__(self):
    return "ScalarDatum(%s)" % ", ".join((
        "step=%r" % (self._step,),
        "wall_time=%r" % (self._wall_time,),
        "value=%r" % (self._value,),
    ))


class RunTagFilter(object):
  """Filters data by run and tag names."""

  def __init__(self, runs=None, tags=None):
    """Construct a `RunTagFilter`.

    A time series passes this filter if both its run *and* its tag are
    included in the corresponding whitelists.

    Order and multiplicity are ignored; `runs` and `tags` are treated as
    sets.

    Args:
      runs: Collection of run names, as strings, or `None` to admit all
        runs.
      tags: Collection of tag names, as strings, or `None` to admit all
        tags.
    """
    self._runs = None if runs is None else frozenset(runs)
    self._tags = None if tags is None else frozenset(tags)

  @property
  def runs(self):
    return self._runs

  @property
  def tags(self):
    return self._tags

  def __repr__(self):
    return "RunTagFilter(%s)" % ", ".join((
        "runs=%r" % (self._runs,),
        "tags=%r" % (self._tags,),
    ))



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Bridge from event multiplexer storage to generic data APIs."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from tensorboard.data import provider
from tensorboard.util import tb_logging
from tensorboard.util import tensor_util


logger = tb_logging.get_logger()


class MultiplexerDataProvider(provider.DataProvider):
  def __init__(self, multiplexer, logdir):
    """Trivial initializer.

    Args:
      multiplexer: A `plugin_event_multiplexer.EventMultiplexer` (note:
        not a boring old `event_multiplexer.EventMultiplexer`).
      logdir: The log directory from which data is being read. Only used
        cosmetically. Should be a `str`.
    """
    self._multiplexer = multiplexer
    self._logdir = logdir

  def _test_run_tag(self, run_tag_filter, run, tag):
    runs = run_tag_filter.runs
    if runs is not None and run not in runs:
      return False
    tags = run_tag_filter.tags
    if tags is not None and tag not in tags:
      return False
    return True

  def _get_first_event_timestamp(self, run_name):
    try:
      return self._multiplexer.FirstEventTimestamp(run_name)
    except ValueError as e:
      return None

  def data_location(self, experiment_id):
    del experiment_id   # ignored
    return str(self._logdir)

  def list_runs(self, experiment_id):
    del experiment_id  # ignored for now
    return [
        provider.Run(
            run_id=run,  # use names as IDs
            run_name=run,
            start_time=self._get_first_event_timestamp(run),
        )
        for run in self._multiplexer.Runs()
    ]

  def list_scalars(self, experiment_id, plugin_name, run_tag_filter=None):
    del experiment_id  # ignored for now
    run_tag_content = self._multiplexer.PluginRunToTagToContent(plugin_name)
    result = {}
    if run_tag_filter is None:
      run_tag_filter = provider.RunTagFilter(runs=None, tags=None)
    for (run, tag_to_content) in six.iteritems(run_tag_content):
      result_for_run = {}
      for tag in tag_to_content:
        if not self._test_run_tag(run_tag_filter, run, tag):
          continue
        result[run] = result_for_run
        max_step = None
        max_wall_time = None
        for event in self._multiplexer.Tensors(run, tag):
          if max_step is None or max_step < event.step:
            max_step = event.step
          if max_wall_time is None or max_wall_time < event.wall_time:
            max_wall_time = event.wall_time
        summary_metadata = self._multiplexer.SummaryMetadata(run, tag)
        result_for_run[tag] = provider.ScalarTimeSeries(
            max_step=max_step,
            max_wall_time=max_wall_time,
            plugin_content=summary_metadata.plugin_data.content,
            description=summary_metadata.summary_description,
            display_name=summary_metadata.display_name,
        )
    return result

  def read_scalars(
      self, experiment_id, plugin_name, downsample=None, run_tag_filter=None
  ):
    # TODO(@wchargin): Downsampling not implemented, as the multiplexer
    # is already downsampled. We could downsample on top of the existing
    # sampling, which would be nice for testing.
    del downsample  # ignored for now
    index = self.list_scalars(
        experiment_id, plugin_name, run_tag_filter=run_tag_filter
    )
    result = {}
    for (run, tags_for_run) in six.iteritems(index):
      result_for_run = {}
      result[run] = result_for_run
      for (tag, metadata) in six.iteritems(tags_for_run):
        events = self._multiplexer.Tensors(run, tag)
        result_for_run[tag] = [self._convert_scalar_event(e) for e in events]
    return result

  def _convert_scalar_event(self, event):
    return provider.ScalarDatum(
        step=event.step,
        wall_time=event.wall_time,
        value=tensor_util.make_ndarray(event.tensor_proto).item(),
    )



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wraps the base_plugin.TBContext to stores additional data shared across
  API handlers for the HParams plugin backend.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os
import threading

import six

from tensorboard.plugins.hparams import api_pb2
from tensorboard.plugins.hparams import metadata
from google.protobuf import json_format
from tensorboard.plugins.scalar import metadata as scalar_metadata


class Context(object):
  """Wraps the base_plugin.TBContext to stores additional data shared across
  API handlers for the HParams plugin backend.

  Before adding fields to this class, carefully consider whether the field
  truelly needs to be accessible to all API handlers or if it can be passed
  separately to the handler constructor.
  We want to avoid this class becoming a magic container of variables that
  have no better place. See http://wiki.c2.com/?MagicContainer
  """

  def __init__(self,
               tb_context,
               max_domain_discrete_len=10):
    """Instantiates a context.

    Args:
      tb_context: base_plugin.TBContext. The "base" context we extend.
      max_domain_discrete_len: int. Only used when computing the experiment
        from the session runs. The maximum number of disticnt values a string
        hyperparameter can have for us to populate its 'domain_discrete' field.
        Typically, only tests should specify a value for this parameter.
    """
    self._tb_context = tb_context
    self._experiment_from_tag = None
    self._experiment_from_tag_lock = threading.Lock()
    self._max_domain_discrete_len = max_domain_discrete_len

  def experiment(self):
    """Returns the experiment protobuffer defining the experiment.

    This method first attempts to find a metadata.EXPERIMENT_TAG tag and
    retrieve the associated protobuffer. If no such tag is found, the method
    will attempt to build a minimal experiment protobuffer by scanning for
    all metadata.SESSION_START_INFO_TAG tags (to compute the hparam_infos
    field of the experiment) and for all scalar tags (to compute the
    metric_infos field of the experiment).

    Returns:
      The experiment protobuffer. If no tags are found from which an experiment
      protobuffer can be built (possibly, because the event data has not been
      completely loaded yet), returns None.
    """
    experiment = self._find_experiment_tag()
    if experiment is None:
      return self._compute_experiment_from_runs()
    return experiment

  @property
  def multiplexer(self):
    return self._tb_context.multiplexer

  @property
  def tb_context(self):
    return self._tb_context

  def _find_experiment_tag(self):
    """Finds the experiment associcated with the metadata.EXPERIMENT_TAG tag.

    Caches the experiment if it was found.

    Returns:
      The experiment or None if no such experiment is found.
    """
    with self._experiment_from_tag_lock:
      if self._experiment_from_tag is None:
        mapping = self.multiplexer.PluginRunToTagToContent(
            metadata.PLUGIN_NAME)
        for tag_to_content in mapping.values():
          if metadata.EXPERIMENT_TAG in tag_to_content:
            self._experiment_from_tag = metadata.parse_experiment_plugin_data(
                tag_to_content[metadata.EXPERIMENT_TAG])
            break
    return self._experiment_from_tag

  def _compute_experiment_from_runs(self):
    """Computes a minimal Experiment protocol buffer by scanning the runs."""
    hparam_infos = self._compute_hparam_infos()
    if not hparam_infos:
      return None
    metric_infos = self._compute_metric_infos()
    return api_pb2.Experiment(hparam_infos=hparam_infos,
                              metric_infos=metric_infos)

  def _compute_hparam_infos(self):
    """Computes a list of api_pb2.HParamInfo from the current run, tag info.

    Finds all the SessionStartInfo messages and collects the hparams values
    appearing in each one. For each hparam attempts to deduce a type that fits
    all its values. Finally, sets the 'domain' of the resulting HParamInfo
    to be discrete if the type is string and the number of distinct values is
    small enough.

    Returns:
      A list of api_pb2.HParamInfo messages.
    """
    run_to_tag_to_content = self.multiplexer.PluginRunToTagToContent(
        metadata.PLUGIN_NAME)
    # Construct a dict mapping an hparam name to its list of values.
    hparams = collections.defaultdict(list)
    for tag_to_content in run_to_tag_to_content.values():
      if metadata.SESSION_START_INFO_TAG not in tag_to_content:
        continue
      start_info = metadata.parse_session_start_info_plugin_data(
          tag_to_content[metadata.SESSION_START_INFO_TAG])
      for (name, value) in six.iteritems(start_info.hparams):
        hparams[name].append(value)

    # Try to construct an HParamInfo for each hparam from its name and list
    # of values.
    result = []
    for (name, values) in six.iteritems(hparams):
      hparam_info = self._compute_hparam_info_from_values(name, values)
      if hparam_info is not None:
        result.append(hparam_info)
    return result

  def _compute_hparam_info_from_values(self, name, values):
    """Builds an HParamInfo message from the hparam name and list of values.

    Args:
      name: string. The hparam name.
      values: list of google.protobuf.Value messages. The list of values for the
        hparam.

    Returns:
      An api_pb2.HParamInfo message.
    """
    # Figure out the type from the values.
    # Ignore values whose type is not listed in api_pb2.DataType
    # If all values have the same type, then that is the type used.
    # Otherwise, the returned type is DATA_TYPE_STRING.
    result = api_pb2.HParamInfo(name=name, type=api_pb2.DATA_TYPE_UNSET)
    distinct_values = set(
        _protobuf_value_to_string(v) for v in values if _protobuf_value_type(v))
    for v in values:
      v_type = _protobuf_value_type(v)
      if not v_type:
        continue
      if result.type == api_pb2.DATA_TYPE_UNSET:
        result.type = v_type
      elif result.type != v_type:
        result.type = api_pb2.DATA_TYPE_STRING
      if result.type == api_pb2.DATA_TYPE_STRING:
        # A string result.type does not change, so we can exit the loop.
        break

    # If we couldn't figure out a type, then we can't compute the hparam_info.
    if result.type == api_pb2.DATA_TYPE_UNSET:
      return None

    # If the result is a string, set the domain to be the distinct values if
    # there aren't too many of them.
    if (result.type == api_pb2.DATA_TYPE_STRING
        and len(distinct_values) <= self._max_domain_discrete_len):
      result.domain_discrete.extend(distinct_values)

    return result

  def _compute_metric_infos(self):
    return (api_pb2.MetricInfo(name=api_pb2.MetricName(group=group, tag=tag))
            for tag, group in self._compute_metric_names())

  def _compute_metric_names(self):
    """Computes the list of metric names from all the scalar (run, tag) pairs.

    The return value is a list of (tag, group) pairs representing the metric
    names. The list is sorted in Python tuple-order (lexicographical).

    For example, if the scalar (run, tag) pairs are:
    ("exp/session1", "loss")
    ("exp/session2", "loss")
    ("exp/session2/eval", "loss")
    ("exp/session2/validation", "accuracy")
    ("exp/no-session", "loss_2"),
    and the runs corresponding to sessions are "exp/session1", "exp/session2",
    this method will return [("loss", ""), ("loss", "/eval"), ("accuracy",
    "/validation")]

    More precisely, each scalar (run, tag) pair is converted to a (tag, group)
    metric name, where group is the suffix of run formed by removing the
    longest prefix which is a session run. If no session run is a prefix of
    'run', the pair is skipped.

    Returns:
      A python list containing pairs. Each pair is a (tag, group) pair
      representing a metric name used in some session.
    """
    session_runs = self._build_session_runs_set()
    metric_names_set = set()
    run_to_tag_to_content = self.multiplexer.PluginRunToTagToContent(
        scalar_metadata.PLUGIN_NAME)
    for (run, tag_to_content) in six.iteritems(run_to_tag_to_content):
      session = _find_longest_parent_path(session_runs, run)
      if not session:
        continue
      group = os.path.relpath(run, session)
      # relpath() returns "." for the 'session' directory, we use an empty
      # string.
      if group == ".":
        group = ""
      metric_names_set.update((tag, group) for tag in tag_to_content.keys())
    metric_names_list = list(metric_names_set)
    # Sort metrics for determinism.
    metric_names_list.sort()
    return metric_names_list

  def _build_session_runs_set(self):
    result = set()
    run_to_tag_to_content = self.multiplexer.PluginRunToTagToContent(
        metadata.PLUGIN_NAME)
    for (run, tag_to_content) in six.iteritems(run_to_tag_to_content):
      if metadata.SESSION_START_INFO_TAG in tag_to_content:
        result.add(run)
    return result


def _find_longest_parent_path(path_set, path):
  """Finds the longest "parent-path" of 'path' in 'path_set'.

  This function takes and returns "path-like" strings which are strings
  made of strings separated by os.sep. No file access is performed here, so
  these strings need not correspond to actual files in some file-system..
  This function returns the longest ancestor path
  For example, for path_set=["/foo/bar", "/foo", "/bar/foo"] and
  path="/foo/bar/sub_dir", returns "/foo/bar".

  Args:
    path_set: set of path-like strings -- e.g. a list of strings separated by
      os.sep. No actual disk-access is performed here, so these need not
      correspond to actual files.
    path: a path-like string.

  Returns:
    The element in path_set which is the longest parent directory of 'path'.
  """
  # This could likely be more efficiently implemented with a trie
  # data-structure, but we don't want to add an extra dependency for that.
  while path not in path_set:
    if not path:
      return None
    path = os.path.dirname(path)
  return path


def _protobuf_value_type(value):
  """Returns the type of the google.protobuf.Value message as an api.DataType.

  Returns None if the type of 'value' is not one of the types supported in
  api_pb2.DataType.

  Args:
    value: google.protobuf.Value message.
  """
  if value.HasField("number_value"):
    return api_pb2.DATA_TYPE_FLOAT64
  if value.HasField("string_value"):
    return api_pb2.DATA_TYPE_STRING
  if value.HasField("bool_value"):
    return api_pb2.DATA_TYPE_BOOL
  return None


def _protobuf_value_to_string(value):
  """Returns a string representation of given google.protobuf.Value message.

  Args:
    value: google.protobuf.Value message. Assumed to be of type 'number',
      'string' or 'bool'.
  """
  value_in_json = json_format.MessageToJson(value)
  if value.HasField("string_value"):
    # Remove the quotations.
    return value_in_json[1:-1]
  return value_in_json



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Constants used in the HParams plugin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.compat.proto import summary_pb2
from tensorboard.plugins.hparams import error
from tensorboard.plugins.hparams import plugin_data_pb2


PLUGIN_NAME = 'hparams'
PLUGIN_DATA_VERSION = 0


EXPERIMENT_TAG = '_hparams_/experiment'
SESSION_START_INFO_TAG = '_hparams_/session_start_info'
SESSION_END_INFO_TAG = '_hparams_/session_end_info'


def create_summary_metadata(hparams_plugin_data_pb):
  """Returns a summary metadata for the HParams plugin.

  Returns a summary_pb2.SummaryMetadata holding a copy of the given
  HParamsPluginData message in its plugin_data.content field.
  Sets the version field of the hparams_plugin_data_pb copy to
  PLUGIN_DATA_VERSION.

  Args:
    hparams_plugin_data_pb: the HParamsPluginData protobuffer to use.
  """
  if not isinstance(hparams_plugin_data_pb, plugin_data_pb2.HParamsPluginData):
    raise TypeError('Needed an instance of plugin_data_pb2.HParamsPluginData.'
                    ' Got: %s' % type(hparams_plugin_data_pb))
  content = plugin_data_pb2.HParamsPluginData()
  content.CopyFrom(hparams_plugin_data_pb)
  content.version = PLUGIN_DATA_VERSION
  return summary_pb2.SummaryMetadata(
      plugin_data=summary_pb2.SummaryMetadata.PluginData(
          plugin_name=PLUGIN_NAME, content=content.SerializeToString()))


def parse_experiment_plugin_data(content):
  """Returns the experiment from HParam's SummaryMetadata.plugin_data.content.

  Raises HParamsError if the content doesn't have 'experiment' set or
  this file is incompatible with the version of the metadata stored.

  Args:
    content: The SummaryMetadata.plugin_data.content to use.
  """
  return _parse_plugin_data_as(content, 'experiment')


def parse_session_start_info_plugin_data(content):
  """Returns session_start_info from the plugin_data.content.

  Raises HParamsError if the content doesn't have 'session_start_info' set or
  this file is incompatible with the version of the metadata stored.

  Args:
    content: The SummaryMetadata.plugin_data.content to use.
  """
  return _parse_plugin_data_as(content, 'session_start_info')


def parse_session_end_info_plugin_data(content):
  """Returns session_end_info from the plugin_data.content.

  Raises HParamsError if the content doesn't have 'session_end_info' set or
  this file is incompatible with the version of the metadata stored.

  Args:
    content: The SummaryMetadata.plugin_data.content to use.
  """
  return _parse_plugin_data_as(content, 'session_end_info')


def _parse_plugin_data_as(content, data_oneof_field):
  """Returns a data oneof's field from plugin_data.content.

  Raises HParamsError if the content doesn't have 'data_oneof_field' set or
  this file is incompatible with the version of the metadata stored.

  Args:
    content: The SummaryMetadata.plugin_data.content to use.
    data_oneof_field: string. The name of the data oneof field to return.
  """
  plugin_data = plugin_data_pb2.HParamsPluginData.FromString(content)
  if plugin_data.version != PLUGIN_DATA_VERSION:
    raise error.HParamsError(
        'Only supports plugin_data version: %s; found: %s in: %s' %
        (PLUGIN_DATA_VERSION, plugin_data.version, plugin_data))
  if not plugin_data.HasField(data_oneof_field):
    raise error.HParamsError(
        'Expected plugin_data.%s to be set. Got: %s' %
        (data_oneof_field, plugin_data))
  return getattr(plugin_data, data_oneof_field)
